#include <msp/gl/extensions/arb_buffer_storage.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_map_buffer_range.h>
#include <msp/gl/extensions/arb_vertex_buffer_object.h>
#include <msp/gl/extensions/khr_debug.h>
#include <msp/gl/extensions/oes_mapbuffer.h>
#include "buffer.h"
#include "buffer_backend.h"
#include "device.h"

using namespace std;

namespace Msp {
namespace GL {

OpenGLBuffer::OpenGLBuffer()
{
	static Require _req(ARB_vertex_buffer_object);

	if(ARB_direct_state_access)
		glCreateBuffers(1, &id);
	else
		glGenBuffers(1, &id);
}

OpenGLBuffer::OpenGLBuffer(OpenGLBuffer &&other):
	id(other.id)
{
	other.id = 0;
}

OpenGLBuffer::~OpenGLBuffer()
{
	if(this==Device::get_current().get_state().scratch_buffer)
		unbind_scratch();
	if(id)
		glDeleteBuffers(1, &id);
}

void OpenGLBuffer::allocate()
{
	size_t size = static_cast<const Buffer *>(this)->size;

	if(ARB_buffer_storage)
	{
		static const int flags = GL_MAP_READ_BIT|GL_MAP_WRITE_BIT|GL_DYNAMIC_STORAGE_BIT;
		if(ARB_direct_state_access)
			glNamedBufferStorage(id, size, 0, flags);
		else
		{
			bind_scratch();
			glBufferStorage(GL_ARRAY_BUFFER, size, 0, flags);
		}
	}
	else if(ARB_direct_state_access)
		glNamedBufferData(id, size, 0, GL_STATIC_DRAW);
	else
	{
		bind_scratch();
		glBufferData(GL_ARRAY_BUFFER, size, 0, GL_STATIC_DRAW);
	}
}

void OpenGLBuffer::sub_data(size_t off, size_t sz, const void *d)
{
	if(ARB_direct_state_access)
		glNamedBufferSubData(id, off, sz, d);
	else
	{
		bind_scratch();
		glBufferSubData(GL_ARRAY_BUFFER, off, sz, d);
	}
}

void *OpenGLBuffer::map()
{
	static Require _req(ARB_map_buffer_range);

	size_t size = static_cast<const Buffer *>(this)->size;

	if(ARB_direct_state_access)
		return glMapNamedBufferRange(id, 0, size, GL_MAP_READ_BIT|GL_MAP_WRITE_BIT);
	else
	{
		bind_scratch();
		void *result = glMapBufferRange(GL_ARRAY_BUFFER, 0, size, GL_MAP_READ_BIT|GL_MAP_WRITE_BIT);
		return result;
	}
}

bool OpenGLBuffer::unmap()
{
	if(ARB_direct_state_access)
		return glUnmapNamedBuffer(id);
	else if(OES_mapbuffer)
	{
		bind_scratch();
		bool result = glUnmapBuffer(GL_ARRAY_BUFFER);
		return result;
	}
	else
		return true;
}

void OpenGLBuffer::set_debug_name(const string &name)
{
#ifdef DEBUG
	if(KHR_debug)
		glObjectLabel(GL_BUFFER, id, name.size(), name.c_str());
#else
	(void)name;
#endif
}

void OpenGLBuffer::bind_scratch()
{
	const OpenGLBuffer *&scratch_binding = Device::get_current().get_state().scratch_buffer;
	if(scratch_binding!=this)
	{
		glBindBuffer(GL_ARRAY_BUFFER, id);
		scratch_binding = this;
	}
}

void OpenGLBuffer::unbind_scratch()
{
	const OpenGLBuffer *&scratch_binding = Device::get_current().get_state().scratch_buffer;
	if(scratch_binding)
	{
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		scratch_binding = 0;
	}
}

} // namespace GL
} // namespace Msp
