#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_sampler_objects.h>
#include <msp/gl/extensions/arb_shader_objects.h>
#include <msp/gl/extensions/arb_uniform_buffer_object.h>
#include <msp/gl/extensions/arb_vertex_array_object.h>
#include <msp/gl/extensions/ext_framebuffer_object.h>
#include <msp/gl/extensions/msp_primitive_restart.h>
#include "blend.h"
#include "buffer.h"
#include "depthtest.h"
#include "device.h"
#include "framebuffer.h"
#include "gl.h"
#include "pipelinestate.h"
#include "pipelinestate_backend.h"
#include "program.h"
#include "rect.h"
#include "sampler.h"
#include "stenciltest.h"
#include "texture.h"
#include "uniformblock.h"
#include "vertexsetup.h"

using namespace std;

namespace Msp {
namespace GL {

OpenGLPipelineState::~OpenGLPipelineState()
{
	if(applied_to)
		applied_to->get_state().last_pipeline = 0;
}

void OpenGLPipelineState::apply() const
{
	const PipelineState &self = *static_cast<const PipelineState *>(this);
	Device &device = Device::get_current();
	unsigned mask = changes;

	if(applied_to && applied_to!=&device)
	{
		applied_to->get_state().last_pipeline = 0;
		mask = ~0U;
	}

	OpenGLDeviceState &dev_state = device.get_state();
	if(!dev_state.last_pipeline)
		OpenGLTexture::unbind_scratch();
	
	if(this!=dev_state.last_pipeline)
	{
		if(dev_state.last_pipeline)
			dev_state.last_pipeline->applied_to = 0;
		mask = ~0U;
	}

	if(mask&PipelineState::FRAMEBUFFER)
	{
		const Framebuffer *framebuffer = self.framebuffer;
		glBindFramebuffer(GL_FRAMEBUFFER, framebuffer ? framebuffer->id : 0);
		if(framebuffer)
		{
			framebuffer->refresh();
			framebuffer->require_complete();
		}
	}

	if(mask&(PipelineState::VIEWPORT|PipelineState::FRAMEBUFFER))
	{
		if(const Rect *viewport = self.viewport)
			glViewport(viewport->left, viewport->bottom, viewport->width, viewport->height);
		else if(const Framebuffer *framebuffer = self.framebuffer)
			glViewport(0, 0, framebuffer->get_width(), framebuffer->get_height());
	}

	if(mask&PipelineState::SCISSOR)
	{
		if(const Rect *scissor = self.scissor)
		{
			glEnable(GL_SCISSOR_TEST);
			glScissor(scissor->left, scissor->bottom, scissor->width, scissor->height);
		}
		else
			glDisable(GL_SCISSOR_TEST);
	}

	if(mask&PipelineState::SHPROG)
	{
		glUseProgram(self.shprog ? self.shprog->id : 0);

		unsigned ncd = (self.shprog ? self.shprog->get_n_clip_distances() : 0);
		if(ncd!=dev_state.n_clip_distances)
		{
			for(unsigned i=0; (i<ncd || i<dev_state.n_clip_distances); ++i)
			{
				if(i<ncd)
					glEnable(GL_CLIP_PLANE0+i);
				else
					glDisable(GL_CLIP_PLANE0+i);
			}
			dev_state.n_clip_distances = ncd;
		}
	}

	if(mask&PipelineState::UNIFORMS)
	{
		for(const PipelineState::BoundUniformBlock &u: self.uniform_blocks)
			if(u.changed || mask==~0U)
			{
				if(u.used)
				{
					if(u.binding>=0)
					{
						glBindBufferRange(GL_UNIFORM_BUFFER, u.binding, u.block->get_buffer()->id, u.block->get_offset(), u.block->get_data_size());
						dev_state.bound_uniform_blocks[u.binding] = 1;
					}
					else if(u.binding==ReflectData::DEFAULT_BLOCK && self.shprog)
					{
						const char *data = static_cast<const char *>(u.block->get_data_pointer());
						for(const Program::UniformCall &call: self.shprog->uniform_calls)
							call.func(call.location, call.size, data+call.location*16);
					}
				}

				u.changed = false;
			}
	}

	if(mask&PipelineState::TEXTURES)
	{
		for(const PipelineState::BoundTexture &t: self.textures)
			if(t.changed || mask==~0U)
			{
				if(t.used)
				{
					if(ARB_direct_state_access)
						glBindTextureUnit(t.binding, t.texture->id);
					else
					{
						glActiveTexture(GL_TEXTURE0+t.binding);
						if(dev_state.bound_tex_targets[t.binding] && static_cast<int>(t.texture->target)!=dev_state.bound_tex_targets[t.binding])
							glBindTexture(dev_state.bound_tex_targets[t.binding], 0);
						glBindTexture(t.texture->target, t.texture->id);
					}

					dev_state.bound_tex_targets[t.binding] = t.texture->target;

					glBindSampler(t.binding, t.sampler->id);
					t.sampler->refresh();
				}

				t.changed = false;
			}
	}

	if(mask&PipelineState::VERTEX_SETUP)
	{
		const VertexSetup *vertex_setup = self.vertex_setup;
		glBindVertexArray(vertex_setup ? vertex_setup->id : 0);
		if(vertex_setup)
		{
			static Require _req(MSP_primitive_restart);

			vertex_setup->refresh();
			unsigned ri = (vertex_setup->get_index_type()==UNSIGNED_INT ? 0xFFFFFFFF : 0xFFFF);
			if(ri!=dev_state.restart_index)
			{
				if(!dev_state.restart_index)
					glEnable(GL_PRIMITIVE_RESTART);
				glPrimitiveRestartIndex(ri);
				dev_state.restart_index = ri;
			}
		}
	}

	if(mask&PipelineState::FACE_CULL)
	{
		glFrontFace(self.front_face==CLOCKWISE ? GL_CW : GL_CCW);

		if(self.face_cull!=NO_CULL && self.front_face!=NON_MANIFOLD)
		{
			glEnable(GL_CULL_FACE);
			glCullFace(self.face_cull==CULL_FRONT ? GL_FRONT : GL_BACK);
		}
		else
			glDisable(GL_CULL_FACE);
	}

	if(mask&PipelineState::DEPTH_TEST)
	{
		const DepthTest *depth_test = self.depth_test;
		if(depth_test && depth_test->enabled)
		{
			glEnable(GL_DEPTH_TEST);
			glDepthFunc(get_gl_predicate(depth_test->compare));
		}
		else
			glDisable(GL_DEPTH_TEST);

		glDepthMask(!depth_test || depth_test->write);
	}

	if(mask&PipelineState::STENCIL_TEST)
	{
		const StencilTest *stencil_test = self.stencil_test;
		if(stencil_test && stencil_test->enabled)
		{
			glEnable(GL_STENCIL_TEST);
			glStencilFunc(get_gl_predicate(stencil_test->compare), stencil_test->reference, 0xFFFFFFFF);
			glStencilOp(get_gl_stencil_op(stencil_test->stencil_fail_op), get_gl_stencil_op(stencil_test->depth_fail_op), get_gl_stencil_op(stencil_test->depth_pass_op));
		}
		else
			glDisable(GL_STENCIL_TEST);
	}

	if(mask&PipelineState::BLEND)
	{
		const Blend *blend = self.blend;
		if(blend && blend->enabled)
		{
			glEnable(GL_BLEND);
			glBlendEquation(get_gl_blend_equation(blend->equation));
			glBlendFunc(get_gl_blend_factor(blend->src_factor), get_gl_blend_factor(blend->dst_factor));
			glBlendColor(blend->constant.r, blend->constant.g, blend->constant.b, blend->constant.a);
			ColorWriteMask cw = blend->write_mask;
			glColorMask((cw&WRITE_RED)!=0, (cw&WRITE_GREEN)!=0, (cw&WRITE_BLUE)!=0, (cw&WRITE_ALPHA)!=0);
		}
		else
		{
			glDisable(GL_BLEND);
			glColorMask(true, true, true, true);
		}
	}

	applied_to = &device;
	dev_state.last_pipeline = this;
	changes = 0;
}

void OpenGLPipelineState::clear()
{
	OpenGLDeviceState &dev_state = Device::get_current().get_state();
	if(dev_state.last_pipeline)
	{
		glUseProgram(0);
		glBindVertexArray(0);

		for(unsigned i=0; i<dev_state.n_clip_distances; ++i)
			glDisable(GL_CLIP_PLANE0+i);
		dev_state.n_clip_distances = 0;

		for(unsigned i=0; i<dev_state.bound_tex_targets.size(); ++i)
			if(dev_state.bound_tex_targets[i])
			{
				if(ARB_direct_state_access)
					glBindTextureUnit(i, 0);
				else
				{
					glActiveTexture(GL_TEXTURE0+i);
					glBindTexture(dev_state.bound_tex_targets[i], 0);
				}
				dev_state.bound_tex_targets[i] = 0;
			}

		for(unsigned i=0; i<dev_state.bound_uniform_blocks.size(); ++i)
			if(dev_state.bound_uniform_blocks[i])
			{
				glBindBufferBase(GL_UNIFORM_BUFFER, i, 0);
				dev_state.bound_uniform_blocks[i] = 0;
			}

		glDisable(GL_DEPTH_TEST);
		glDepthMask(true);
		glDisable(GL_STENCIL_TEST);
		glDisable(GL_BLEND);

		dev_state.last_pipeline->applied_to = 0;
		dev_state.last_pipeline = 0;
	}
}

} // namespace GL
} // namespace Msp
