#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_sampler_objects.h>
#include <msp/gl/extensions/arb_shadow.h>
#include <msp/gl/extensions/ext_texture_filter_anisotropic.h>
#include <msp/gl/extensions/ext_texture3d.h>
#include <msp/gl/extensions/khr_debug.h>
#include "gl.h"
#include "sampler.h"
#include "sampler_backend.h"

using namespace std;

namespace Msp {
namespace GL {

OpenGLSampler::OpenGLSampler()
{
	static Require _req(ARB_sampler_objects);
	static Require _req2(EXT_texture3D);
	static Require _req3(ARB_shadow);

	if(ARB_direct_state_access)
		glCreateSamplers(1, &id);
	else
		glGenSamplers(1, &id);
}

OpenGLSampler::OpenGLSampler(OpenGLSampler &&other):
	id(other.id)
{
	other.id = 0;
}

OpenGLSampler::~OpenGLSampler()
{
	if(id)
		glDeleteSamplers(1, &id);
}

bool OpenGLSampler::check_anisotropic(bool require)
{
	if(require)
		static Require _req(EXT_texture_filter_anisotropic);
	return EXT_texture_filter_anisotropic;
}

void OpenGLSampler::update(unsigned mask) const
{
	const Sampler &self = *static_cast<const Sampler *>(this);
	if(mask&Sampler::MIN_FILTER)
		glSamplerParameteri(id, GL_TEXTURE_MIN_FILTER, get_gl_filter(self.min_filter));
	if(mask&Sampler::MAG_FILTER)
		glSamplerParameteri(id, GL_TEXTURE_MAG_FILTER, get_gl_filter(self.mag_filter));
	if(mask&Sampler::MAX_ANISOTROPY)
		glSamplerParameterf(id, GL_TEXTURE_MAX_ANISOTROPY_EXT, self.max_anisotropy);
	if(mask&Sampler::WRAP_S)
		glSamplerParameteri(id, GL_TEXTURE_WRAP_S, get_gl_wrap(self.wrap_s));
	if(mask&Sampler::WRAP_T)
		glSamplerParameteri(id, GL_TEXTURE_WRAP_T, get_gl_wrap(self.wrap_t));
	if(mask&Sampler::WRAP_R)
		glSamplerParameteri(id, GL_TEXTURE_WRAP_R, get_gl_wrap(self.wrap_r));
	if(mask&Sampler::BORDER_COLOR)
		glSamplerParameterfv(id, GL_TEXTURE_BORDER_COLOR, &self.border_color.r);
	if(mask&Sampler::COMPARE)
	{
		glSamplerParameteri(id, GL_TEXTURE_COMPARE_MODE, (self.compare ? GL_COMPARE_R_TO_TEXTURE : GL_NONE));
		if(self.compare)
			glSamplerParameteri(id, GL_TEXTURE_COMPARE_FUNC, get_gl_predicate(self.cmp_func));
	}
}

void OpenGLSampler::set_debug_name(const string &name)
{
#ifdef DEBUG
	if(KHR_debug)
		glObjectLabel(GL_SAMPLER, id, name.size(), name.c_str());
#else
	(void)name;
#endif
}


unsigned get_gl_filter(unsigned filter)
{
	switch(filter)
	{
	case NEAREST: return GL_NEAREST;
	case LINEAR: return GL_LINEAR;
	case NEAREST_MIPMAP_NEAREST: return GL_NEAREST_MIPMAP_NEAREST;
	case NEAREST_MIPMAP_LINEAR: return GL_NEAREST_MIPMAP_LINEAR;
	case LINEAR_MIPMAP_NEAREST: return GL_LINEAR_MIPMAP_NEAREST;
	case LINEAR_MIPMAP_LINEAR: return GL_LINEAR_MIPMAP_LINEAR;
	default: throw invalid_argument("get_gl_filter");
	}
}

unsigned get_gl_wrap(unsigned wrap)
{
	switch(wrap)
	{
	case REPEAT: return GL_REPEAT;
	case CLAMP_TO_EDGE: return GL_CLAMP_TO_EDGE;
	case CLAMP_TO_BORDER: return GL_CLAMP_TO_BORDER;
	case MIRRORED_REPEAT: return GL_MIRRORED_REPEAT;
	default: throw invalid_argument("get_gl_wrap");
	}
}

} // namespace GL
} // namespace Msp
