#include <msp/datafile/rawdata.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_texture_storage.h>
#include <msp/gl/extensions/arb_vertex_buffer_object.h>
#include <msp/graphics/imageloader.h>
#include "buffer.h"
#include "gl.h"
#include "texture2d.h"
#include "texture2d_backend.h"

using namespace std;

namespace Msp {
namespace GL {

class OpenGLTexture2D::AsyncLoader: public Resource::AsyncLoader
{
private:
	Texture2D &texture;
	IO::Seekable &io;
	Buffer pixel_buffer;
	char *mapped_address = 0;
	Graphics::Image image;
	Graphics::ImageLoader *img_loader = 0;
	DataFile::RawData *raw_data = 0;
	unsigned n_bytes = 0;
	int phase = 0;

public:
	AsyncLoader(Texture2D &, IO::Seekable &);
	~AsyncLoader();

	virtual bool needs_sync() const;
	virtual bool process();
};


OpenGLTexture2D::OpenGLTexture2D():
	Texture(GL_TEXTURE_2D)
{ }

void OpenGLTexture2D::allocate()
{
	const Texture2D &self = *static_cast<const Texture2D *>(this);

	if(!id)
		create();

	GLenum gl_fmt = get_gl_pixelformat(storage_fmt);
	if(ARB_texture_storage)
	{
		if(ARB_direct_state_access)
			glTextureStorage2D(id, self.levels, gl_fmt, self.width, self.height);
		else
		{
			bind_scratch();
			glTexStorage2D(target, self.levels, gl_fmt, self.width, self.height);
		}
	}
	else
	{
		bind_scratch();
		glTexParameteri(target, GL_TEXTURE_MAX_LEVEL, self.levels-1);
		GLenum comp = get_gl_components(get_components(storage_fmt));
		GLenum type = get_gl_type(get_component_type(storage_fmt));
		for(unsigned i=0; i<self.levels; ++i)
		{
			auto lv_size = self.get_level_size(i);
			glTexImage2D(target, i, gl_fmt, lv_size.x, lv_size.y, 0, comp, type, 0);
		}
	}

	apply_swizzle();
}

void OpenGLTexture2D::sub_image(unsigned level, int x, int y, unsigned wd, unsigned ht, const void *data)
{
	GLenum comp = get_gl_components(get_components(storage_fmt));
	GLenum type = get_gl_type(get_component_type(storage_fmt));
	if(ARB_direct_state_access)
		glTextureSubImage2D(id, level, x, y, wd, ht, comp, type, data);
	else
	{
		bind_scratch();
		glTexSubImage2D(target, level, x, y, wd, ht, comp, type, data);
	}
}

void OpenGLTexture2D::sub_image(unsigned level, int x, int y, unsigned wd, unsigned ht, const Buffer &buffer, unsigned offset)
{
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER, buffer.id);
	sub_image(level, x, y, wd, ht, reinterpret_cast<void *>(offset));
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
}

Resource::AsyncLoader *OpenGLTexture2D::load(IO::Seekable &io, const Resources *)
{
	return new AsyncLoader(static_cast<Texture2D &>(*this), io);
}

uint64_t OpenGLTexture2D::get_data_size() const
{
	if(!id)
		return 0;

	const Texture2D &self = *static_cast<const Texture2D *>(this);

	size_t level_size = self.width*self.height*get_pixel_size(format);
	size_t total_size = level_size;
	for(unsigned i=0; i<self.levels; ++i, level_size>>=2)
		total_size += level_size;
	return total_size;
}

void OpenGLTexture2D::unload()
{
	glDeleteTextures(1, &id);
	id = 0;
}


OpenGLTexture2D::AsyncLoader::AsyncLoader(Texture2D &t, IO::Seekable &i):
	texture(t),
	io(i)
{
	char magic[4] = { };
	io.read(magic, 4);
	io.seek(0, IO::S_BEG);

	if(DataFile::RawData::detect_signature(string(magic, 4)))
	{
		raw_data = new DataFile::RawData;
		raw_data->open_io(io, "async");
	}
	else
		img_loader = Graphics::ImageLoader::open_io(io);
}

OpenGLTexture2D::AsyncLoader::~AsyncLoader()
{
	if(mapped_address)
		pixel_buffer.unmap();
	delete img_loader;
	delete raw_data;
}

bool OpenGLTexture2D::AsyncLoader::needs_sync() const
{
	return phase%2;
}

bool OpenGLTexture2D::AsyncLoader::process()
{
	if(phase==0)
	{
		if(raw_data)
			n_bytes = raw_data->get_size();
		else
		{
			image.load_headers(*img_loader);
			n_bytes = image.get_stride()*image.get_height();
		}
	}
	else if(phase==1)
	{
		pixel_buffer.storage(n_bytes, STREAMING);
		mapped_address = reinterpret_cast<char *>(pixel_buffer.map());
	}
	else if(phase==2)
	{
		if(raw_data)
			raw_data->load_into(mapped_address);
		else
			image.load_into(*img_loader, mapped_address);
	}
	else if(phase==3)
	{
		mapped_address = 0;
		if(!pixel_buffer.unmap())
		{
			phase = 1;
			return false;
		}

		if(!texture.id)
			texture.create();

		if(img_loader)
		{
			unsigned w = image.get_width();
			unsigned h = image.get_height();
			texture.storage(pixelformat_from_image(image, texture.use_srgb_format), w, h);
		}
		texture.OpenGLTexture2D::sub_image(0, 0, 0, texture.width, texture.height, pixel_buffer, 0);

		if(texture.auto_gen_mipmap)
			texture.generate_mipmap();
	}

	++phase;
	return phase>3;
}

} // namespace GL
} // namespace Msp
