#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_texture_swizzle.h>
#include <msp/gl/extensions/ext_framebuffer_object.h>
#include <msp/gl/extensions/khr_debug.h>
#include "device.h"
#include "gl.h"
#include "error.h"
#include "texture.h"
#include "texture_backend.h"

using namespace std;

namespace Msp {
namespace GL {

OpenGLTexture::OpenGLTexture(unsigned t):
	target(t)
{
	static bool alignment_init = false;
	if(!alignment_init)
	{
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		alignment_init = true;
	}
}

OpenGLTexture::OpenGLTexture(OpenGLTexture &&other):
	id(other.id),
	target(other.target),
	debug_name(move(other.debug_name))
{
	other.id = 0;
}

OpenGLTexture::~OpenGLTexture()
{
	if(this==Device::get_current().get_state().scratch_texture)
		unbind_scratch();
	if(id)
		glDeleteTextures(1, &id);
}

void OpenGLTexture::create()
{
	if(id)
		throw invalid_operation("OpenGLTexture::create");
	if(ARB_direct_state_access)
		glCreateTextures(target, 1, &id);
	else
		glGenTextures(1, &id);

#ifdef DEBUG
	if(!debug_name.empty() && KHR_debug)
		glObjectLabel(GL_TEXTURE, id, debug_name.size(), debug_name.c_str());
#endif
}

void OpenGLTexture::require_swizzle()
{
	static Require _req(ARB_texture_swizzle);
}

void OpenGLTexture::apply_swizzle()
{
	ComponentSwizzle swizzle = static_cast<const Texture *>(this)->swizzle;
	if(swizzle==NO_SWIZZLE)
		return;

	const int *swizzle_order = get_gl_swizzle(swizzle);
	if(get_backend_api()==OPENGL_ES)
	{
		set_parameter_i(GL_TEXTURE_SWIZZLE_R, swizzle_order[0]);
		set_parameter_i(GL_TEXTURE_SWIZZLE_G, swizzle_order[1]);
		set_parameter_i(GL_TEXTURE_SWIZZLE_B, swizzle_order[2]);
		set_parameter_i(GL_TEXTURE_SWIZZLE_A, swizzle_order[3]);
	}
	else
	{
		if(ARB_direct_state_access)
			glTextureParameteriv(id, GL_TEXTURE_SWIZZLE_RGBA, swizzle_order);
		else
			glTexParameteriv(target, GL_TEXTURE_SWIZZLE_RGBA, swizzle_order);
	}
}

void OpenGLTexture::set_parameter_i(unsigned param, int value) const
{
	if(ARB_direct_state_access)
		glTextureParameteri(id, param, value);
	else
		glTexParameteri(target, param, value);
}

void OpenGLTexture::generate_mipmap()
{
	// glGenerateMipmap is defined here
	static Require _req(EXT_framebuffer_object);

	if(ARB_direct_state_access)
		glGenerateTextureMipmap(id);
	else
	{
		bind_scratch();
		glGenerateMipmap(target);
	}
}

void OpenGLTexture::set_debug_name(const string &name)
{
#ifdef DEBUG
	debug_name = name;
	if(id && KHR_debug)
		glObjectLabel(GL_TEXTURE, id, name.size(), name.c_str());
#else
	(void)name;
#endif
}

void OpenGLTexture::bind_scratch()
{
	const OpenGLTexture *&scratch_binding = Device::get_current().get_state().scratch_texture;
	if(!scratch_binding)
		glActiveTexture(GL_TEXTURE0);
	if(scratch_binding!=this)
	{
		if(scratch_binding && scratch_binding->target!=target)
			glBindTexture(scratch_binding->target, 0);
		glBindTexture(target, id);
		scratch_binding = this;
	}
}

void OpenGLTexture::unbind_scratch()
{
	const OpenGLTexture *&scratch_binding = Device::get_current().get_state().scratch_texture;
	if(scratch_binding)
	{
		glBindTexture(scratch_binding->target, 0);
		scratch_binding = 0;
	}
}

} // namespace GL
} // namespace Msp
