#include <msp/core/hash.h>
#include <msp/graphics/vulkancontext_platform.h>
#include "batch.h"
#include "commands_backend.h"
#include "device.h"
#include "error.h"
#include "framebuffer.h"
#include "frameformat.h"
#include "pipelinestate.h"
#include "rect.h"
#include "semaphore.h"
#include "swapchaintexture.h"
#include "vulkan.h"

using namespace std;

namespace Msp {
namespace GL {

VulkanCommands::VulkanCommands():
	device(Device::get_current())
{ }

VulkanCommands::~VulkanCommands()
{
	const VulkanFunctions &vk = device.get_functions();

	vk.QueueWaitIdle();
}

void VulkanCommands::begin_buffer()
{
	if(!current_pool)
		throw invalid_operation("VulkanCommands::begin_buffer");

	const VulkanFunctions &vk = device.get_functions();

	if(!current_pool->in_use)
	{
		current_pool->fence.reset();
		current_pool->in_use = true;
	}

	if(current_pool->next_buffer>=current_pool->buffers.size())
	{
		VkCommandBufferAllocateInfo alloc_info = { };
		alloc_info.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
		alloc_info.commandPool = handle_cast<::VkCommandPool>(current_pool->pool);
		alloc_info.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
		alloc_info.commandBufferCount = 1;

		VkCommandBuffer buffer;
		vk.AllocateCommandBuffers(alloc_info, &buffer);
		current_pool->buffers.push_back(buffer);
	}

	current_buffer = current_pool->buffers[current_pool->next_buffer++];

	VkCommandBufferBeginInfo begin_info = { };
	begin_info.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
	begin_info.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;

	vk.BeginCommandBuffer(current_buffer, begin_info);
}

void VulkanCommands::begin_render_pass(bool clear, const ClearValue *clear_values)
{
	framebuffer = pipeline_state->get_framebuffer();
	if(!framebuffer)
		throw invalid_operation("VulkanCommands::begin_render_pass");

	const VulkanFunctions &vk = device.get_functions();

	if(!current_buffer)
		begin_buffer();

	device.get_transfer_queue().dispatch_transfers(current_buffer);

	Synchronizer &sync = device.get_synchronizer();
	sync.reset();

	bool to_present = false;
	unsigned n_attachments = framebuffer->get_format().size();
	for(unsigned i=0; i<n_attachments; ++i)
		if(dynamic_cast<const SwapChainTexture *>(framebuffer->get_attachment(i)))
			to_present = true;
	if(!to_present)
		framebuffer->synchronize(clear);
	VkRenderPass render_pass = device.get_pipeline_cache().get_render_pass(framebuffer->get_format(), clear, !clear_values, to_present);

	framebuffer->refresh();

	sync.barrier(current_buffer);

	VkRenderPassBeginInfo begin_info = { };
	begin_info.sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
	begin_info.renderPass = handle_cast<::VkRenderPass>(render_pass);
	begin_info.framebuffer = handle_cast<::VkFramebuffer>(framebuffer->handle);

	viewport = pipeline_state->get_viewport();
	if(viewport)
	{
		begin_info.renderArea.offset.x = viewport->left;
		begin_info.renderArea.offset.y = viewport->bottom;
		begin_info.renderArea.extent.width = viewport->width;
		begin_info.renderArea.extent.height = viewport->height;
	}
	else
	{
		begin_info.renderArea.extent.width = framebuffer->get_width();
		begin_info.renderArea.extent.height = framebuffer->get_height();
	}

	VkClearValue vk_clear_values[7];
	if(clear_values)
	{
		unsigned i = 0;
		for(FrameAttachment a: framebuffer->get_format())
		{
			if(get_attach_point(a)==get_attach_point(DEPTH_ATTACHMENT))
				vk_clear_values[i].depthStencil.depth = clear_values[i].depth_stencil.depth;
			else if(get_attach_point(a)==get_attach_point(STENCIL_ATTACHMENT))
				vk_clear_values[i].depthStencil.stencil = clear_values[i].depth_stencil.stencil;
			else
			{
				vk_clear_values[i].color.float32[0] = clear_values[i].color.r;
				vk_clear_values[i].color.float32[1] = clear_values[i].color.g;
				vk_clear_values[i].color.float32[2] = clear_values[i].color.b;
				vk_clear_values[i].color.float32[3] = clear_values[i].color.a;
			}
			++i;
		}

		begin_info.clearValueCount = framebuffer->get_format().size();
		begin_info.pClearValues = vk_clear_values;
	}

	vk.CmdBeginRenderPass(current_buffer, begin_info, VK_SUBPASS_CONTENTS_INLINE);
}

void VulkanCommands::end_render_pass()
{
	const VulkanFunctions &vk = device.get_functions();

	vk.CmdEndRenderPass(current_buffer);
	framebuffer = 0;
	viewport = 0;
}

void VulkanCommands::begin_frame(unsigned index)
{
	const VulkanFunctions &vk = device.get_functions();

	unsigned pool_index = index%device.get_n_frames_in_flight();
	if(pool_index>=command_pools.size())
	{
		command_pools.reserve(pool_index+1);
		for(unsigned i=command_pools.size(); i<pool_index+1; ++i)
			command_pools.emplace_back(device);
	}

	current_pool = &command_pools[pool_index];
	if(current_pool->in_use)
	{
		current_pool->fence.wait();
		vk.ResetCommandPool(current_pool->pool, 0);
		current_pool->in_use = false;
		current_pool->next_buffer = 0;
	}
}

void VulkanCommands::submit_frame(Semaphore *wait_sem, Semaphore *signal_sem)
{
	if(!current_buffer)
		return;

	const VulkanFunctions &vk = device.get_functions();
	::VkSemaphore vk_wait_sem = (wait_sem ? handle_cast<::VkSemaphore>(wait_sem->handle) : 0);
	::VkSemaphore vk_signal_sem = (signal_sem ? handle_cast<::VkSemaphore>(signal_sem->handle) : 0);
	VkPipelineStageFlags wait_stages = VK_PIPELINE_STAGE_ALL_GRAPHICS_BIT;

	if(framebuffer)
		end_render_pass();

	vk.EndCommandBuffer(current_buffer);

	VkSubmitInfo submit_info = { };
	submit_info.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
	submit_info.waitSemaphoreCount = (wait_sem ? 1 : 0);
	submit_info.pWaitSemaphores = &vk_wait_sem;
	submit_info.pWaitDstStageMask = &wait_stages;
	submit_info.commandBufferCount = 1;
	submit_info.pCommandBuffers = handle_cast<::VkCommandBuffer *>(&current_buffer);
	submit_info.signalSemaphoreCount = (signal_sem ? 1 : 0);
	submit_info.pSignalSemaphores = &vk_signal_sem;

	vk.QueueSubmit(1, &submit_info, current_pool->fence.handle);

	current_buffer = 0;
}

void VulkanCommands::use_pipeline(const PipelineState *ps)
{
	if(!pipeline_state || !ps || ps->get_framebuffer()!=framebuffer || ps->get_viewport()!=viewport)
		if(framebuffer)
			end_render_pass();

	pipeline_state = ps;
}

void VulkanCommands::clear(const ClearValue *values)
{
	if(framebuffer)
		throw invalid_operation("VulkanCommands::clear");

	begin_render_pass(true, values);
}

void VulkanCommands::draw(const Batch &batch)
{
	draw_instanced(batch, 1);
}

void VulkanCommands::draw_instanced(const Batch &batch, unsigned count)
{
	if(!pipeline_state)
		throw invalid_operation("VulkanCommands::draw_instanced");

	const VulkanFunctions &vk = device.get_functions();

	if(!framebuffer)
		 begin_render_pass(false, 0);

	pipeline_state->refresh();
	pipeline_state->apply(current_buffer);
	unsigned first_index = batch.get_offset()/batch.get_index_size();
	vk.CmdDrawIndexed(current_buffer, batch.size(), count, first_index, 0, 0);
}

void VulkanCommands::resolve_multisample(Framebuffer &)
{
	throw logic_error("VulkanCommands::resolve_multisample is unimplemented");
}

void VulkanCommands::begin_query(const QueryPool &, unsigned)
{
	throw logic_error("VulkanCommands::begin_query is unimplemented");
}

void VulkanCommands::end_query(const QueryPool &, unsigned)
{
	throw logic_error("VulkanCommands::end_query is unimplemented");
}


VulkanCommands::CommandPool::CommandPool(Device &d):
	device(d),
	fence(true)
{
	const VulkanFunctions &vk = device.get_functions();

	VkCommandPoolCreateInfo pool_info = { };
	pool_info.sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
	pool_info.flags = VK_COMMAND_POOL_CREATE_TRANSIENT_BIT;
	pool_info.queueFamilyIndex = device.get_context().get_private().graphics_queue_family;

	vk.CreateCommandPool(pool_info, pool);
}

VulkanCommands::CommandPool::CommandPool(CommandPool &&other):
	device(other.device),
	pool(other.pool),
	fence(move(other.fence)),
	in_use(other.in_use)
{
	other.pool = 0;
}

VulkanCommands::CommandPool::~CommandPool()
{
	const VulkanFunctions &vk = device.get_functions();

	if(pool)
		vk.DestroyCommandPool(pool);
}

} // namespace GL
} // namespace Msp
