#include <msp/core/hash.h>
#include "batch.h"
#include "blend.h"
#include "buffer.h"
#include "depthtest.h"
#include "device.h"
#include "framebuffer.h"
#include "pipelinestate.h"
#include "pipelinestate_backend.h"
#include "program.h"
#include "rect.h"
#include "sampler.h"
#include "stenciltest.h"
#include "structurebuilder.h"
#include "texture.h"
#include "uniformblock.h"
#include "vertexsetup.h"
#include "vulkan.h"

using namespace std;

namespace Msp {
namespace GL {

VulkanPipelineState::VulkanPipelineState():
	device(Device::get_current())
{ }

VulkanPipelineState::VulkanPipelineState(VulkanPipelineState &&other):
	device(other.device),
	handle(other.handle)
{ }

void VulkanPipelineState::update() const
{
	const PipelineState &self = *static_cast<const PipelineState *>(this);

	constexpr unsigned pipeline_mask = PipelineState::SHPROG|PipelineState::VERTEX_SETUP|PipelineState::FACE_CULL|
		PipelineState::DEPTH_TEST|PipelineState::STENCIL_TEST|PipelineState::BLEND|PipelineState::PRIMITIVE_TYPE;
	if(self.changes&pipeline_mask)
		handle = device.get_pipeline_cache().get_pipeline(self);

	if(self.changes&(PipelineState::SHPROG|PipelineState::UNIFORMS|PipelineState::TEXTURES))
	{
		unsigned changed_sets = (self.changes&PipelineState::SHPROG ? ~0U : 0U);
		for(const PipelineState::BoundUniformBlock &u: self.uniform_blocks)
			if(u.changed || changed_sets==~0U)
			{
				u.used = self.shprog->uses_binding(u.binding);
				if(u.binding>=0)
					changed_sets |= 1<<(u.binding>>20);
				u.changed = false;
			}
		for(const PipelineState::BoundTexture &t: self.textures)
			if(t.changed || changed_sets==~0U)
			{
				t.used = self.shprog->uses_binding(t.binding);
				changed_sets |= 1<<(t.binding>>20);
				if(t.sampler)
					t.sampler->refresh();
				t.changed = false;
			}

		descriptor_set_handles.resize(self.shprog->get_n_descriptor_sets());
		for(unsigned i=0; i<descriptor_set_handles.size(); ++i)
			if(changed_sets&(1<<i))
				descriptor_set_handles[i] = device.get_pipeline_cache().get_descriptor_set(self, i);
	}

	self.changes = 0;
}

uint64_t VulkanPipelineState::compute_hash() const
{
	const PipelineState &self = *static_cast<const PipelineState *>(this);
	const FrameFormat &format = self.framebuffer->get_format();

	uint64_t result = hash<64>(self.shprog);
	result = hash_update<64>(result, self.vertex_setup->compute_hash());
	result = hash_round<64>(result, self.primitive_type);

	if(self.front_face!=NON_MANIFOLD && self.face_cull!=NO_CULL)
	{
		result = hash_round<64>(result, self.front_face);
		result = hash_round<64>(result, self.face_cull);
	}

	result = hash_round<64>(result, format.get_samples());

	if(const DepthTest *depth_test = self.depth_test)
		if(depth_test->enabled)
		{
			result = hash_round<64>(result, depth_test->compare);
			result = hash_update<64>(result, depth_test->write);
		}

	if(const StencilTest *stencil_test = self.stencil_test)
		if(stencil_test->enabled)
		{
			result = hash_round<64>(result, stencil_test->compare);
			result = hash_round<64>(result, stencil_test->stencil_fail_op);
			result = hash_round<64>(result, stencil_test->depth_fail_op);
			result = hash_round<64>(result, stencil_test->depth_pass_op);
			result = hash_update<64>(result, stencil_test->reference);
		}

	if(const Blend *blend = self.blend)
		if(blend->enabled)
		{
			result = hash_round<64>(result, blend->equation);
			result = hash_round<64>(result, blend->src_factor);
			result = hash_round<64>(result, blend->dst_factor);
			result = hash_round<64>(result, blend->write_mask);
		}

	for(FrameAttachment a: format)
		result = hash_update<64>(result, a);

	return result;
}

void VulkanPipelineState::fill_creation_info(vector<char> &buffer) const
{
	const PipelineState &self = *static_cast<const PipelineState *>(this);

	const FrameFormat &format = self.framebuffer->get_format();
	VkRenderPass render_pass = device.get_pipeline_cache().get_render_pass(format, false, false, false);

	unsigned n_color_attachments = 0;
	for(FrameAttachment a: format)
	{
		unsigned attach_pt = get_attach_point(a);
		if(attach_pt!=get_attach_point(DEPTH_ATTACHMENT) && attach_pt!=get_attach_point(STENCIL_ATTACHMENT))
			++n_color_attachments;
	}

	StructureBuilder sb(buffer, 10);
	VkGraphicsPipelineCreateInfo *&pipeline_info = sb.add<VkGraphicsPipelineCreateInfo>();
	VkPipelineInputAssemblyStateCreateInfo *&input_assembly_info = sb.add<VkPipelineInputAssemblyStateCreateInfo>();
	VkPipelineViewportStateCreateInfo *&viewport_info = sb.add<VkPipelineViewportStateCreateInfo>();
	VkPipelineRasterizationStateCreateInfo *&raster_info = sb.add<VkPipelineRasterizationStateCreateInfo>();
	VkPipelineMultisampleStateCreateInfo *&multisample_info = sb.add<VkPipelineMultisampleStateCreateInfo>();
	VkPipelineDepthStencilStateCreateInfo *&depth_stencil_info = sb.add<VkPipelineDepthStencilStateCreateInfo>();
	VkPipelineColorBlendStateCreateInfo *&blend_info = sb.add<VkPipelineColorBlendStateCreateInfo>();
	VkPipelineColorBlendAttachmentState *&blend_attachments = sb.add<VkPipelineColorBlendAttachmentState>(n_color_attachments);
	VkPipelineDynamicStateCreateInfo *&dynamic_info = sb.add<VkPipelineDynamicStateCreateInfo>();
	VkDynamicState *&dynamic_states = sb.add<VkDynamicState>(2);

	input_assembly_info->sType = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
	input_assembly_info->topology = static_cast<VkPrimitiveTopology>(get_vulkan_primitive_type(self.primitive_type));
	input_assembly_info->primitiveRestartEnable = true;

	viewport_info->sType = VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
	viewport_info->viewportCount = 1;
	viewport_info->pViewports = 0;
	viewport_info->scissorCount = 1;
	viewport_info->pScissors = 0;

	raster_info->sType = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
	raster_info->depthClampEnable = VK_FALSE;
	raster_info->rasterizerDiscardEnable = VK_FALSE;
	raster_info->polygonMode = VK_POLYGON_MODE_FILL;
	if(self.face_cull==NO_CULL || self.front_face==NON_MANIFOLD)
	{
		raster_info->cullMode = VK_CULL_MODE_NONE;
		raster_info->frontFace = VK_FRONT_FACE_COUNTER_CLOCKWISE;
	}
	else
	{
		raster_info->cullMode = (self.face_cull==CULL_FRONT ? VK_CULL_MODE_FRONT_BIT : VK_CULL_MODE_BACK_BIT);
		raster_info->frontFace = (self.front_face==CLOCKWISE ? VK_FRONT_FACE_CLOCKWISE : VK_FRONT_FACE_COUNTER_CLOCKWISE);
	}
	raster_info->depthBiasEnable = VK_FALSE;
	raster_info->depthBiasConstantFactor = 0.0f;
	raster_info->depthBiasClamp = 0.0f;
	raster_info->depthBiasSlopeFactor = 0.0f;
	raster_info->lineWidth = 1.0f;

	multisample_info->sType = VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
	multisample_info->rasterizationSamples = static_cast<VkSampleCountFlagBits>(get_vulkan_samples(format.get_samples()));
	multisample_info->sampleShadingEnable = VK_FALSE;
	multisample_info->minSampleShading = 1.0f;
	multisample_info->pSampleMask = 0;
	multisample_info->alphaToCoverageEnable = VK_FALSE;
	multisample_info->alphaToOneEnable = VK_FALSE;

	depth_stencil_info->sType = VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO;
	if(const DepthTest *depth_test = self.depth_test)
	{
		depth_stencil_info->depthTestEnable = depth_test->enabled;
		depth_stencil_info->depthWriteEnable = depth_test->write;
		depth_stencil_info->depthCompareOp = static_cast<VkCompareOp>(get_vulkan_predicate(depth_test->compare));
		depth_stencil_info->depthBoundsTestEnable = VK_FALSE;
	}
	if(const StencilTest *stencil_test = self.stencil_test)
	{
		depth_stencil_info->stencilTestEnable = stencil_test->enabled;
		depth_stencil_info->front.failOp = static_cast<VkStencilOp>(get_vulkan_stencil_op(stencil_test->stencil_fail_op));
		depth_stencil_info->front.passOp = static_cast<VkStencilOp>(get_vulkan_stencil_op(stencil_test->depth_pass_op));
		depth_stencil_info->front.depthFailOp = static_cast<VkStencilOp>(get_vulkan_stencil_op(stencil_test->depth_fail_op));
		depth_stencil_info->front.compareOp = static_cast<VkCompareOp>(get_vulkan_predicate(stencil_test->compare));
		depth_stencil_info->front.compareMask = 0xFFFFFFFFU;
		depth_stencil_info->front.writeMask = 0xFFFFFFFFU;
		depth_stencil_info->front.reference = stencil_test->reference;
		depth_stencil_info->back = depth_stencil_info->front;
	}

	if(const Blend *blend = self.blend)
	{
		for(unsigned i=0; i<n_color_attachments; ++i)
		{
			blend_attachments[i].blendEnable = blend->enabled;
			blend_attachments[i].srcColorBlendFactor = static_cast<VkBlendFactor>(get_vulkan_blend_factor(blend->src_factor));
			blend_attachments[i].dstColorBlendFactor = static_cast<VkBlendFactor>(get_vulkan_blend_factor(blend->dst_factor));
			blend_attachments[i].colorBlendOp = static_cast<VkBlendOp>(get_vulkan_blend_equation(blend->equation));
			blend_attachments[i].srcAlphaBlendFactor = blend_attachments[i].srcColorBlendFactor;
			blend_attachments[i].dstAlphaBlendFactor = blend_attachments[i].dstColorBlendFactor;
			blend_attachments[i].alphaBlendOp = blend_attachments[i].colorBlendOp;
			blend_attachments[i].colorWriteMask = get_vulkan_color_mask(blend->write_mask);
		}
	}
	else
	{
		for(unsigned i=0; i<n_color_attachments; ++i)
			blend_attachments[i].colorWriteMask = VK_COLOR_COMPONENT_R_BIT|VK_COLOR_COMPONENT_G_BIT|VK_COLOR_COMPONENT_B_BIT|VK_COLOR_COMPONENT_A_BIT;
	}

	blend_info->sType = VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
	blend_info->attachmentCount = n_color_attachments;
	blend_info->pAttachments = blend_attachments;

	dynamic_states[0] = VK_DYNAMIC_STATE_VIEWPORT;
	dynamic_states[1] = VK_DYNAMIC_STATE_SCISSOR;

	dynamic_info->sType = VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO;
	dynamic_info->dynamicStateCount = 2;
	dynamic_info->pDynamicStates = dynamic_states;

	pipeline_info->sType = VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;

	pipeline_info->pInputAssemblyState = input_assembly_info;
	pipeline_info->pTessellationState = 0;
	pipeline_info->pViewportState = viewport_info;
	pipeline_info->pRasterizationState = raster_info;
	pipeline_info->pMultisampleState = multisample_info;
	pipeline_info->pDepthStencilState = depth_stencil_info;
	pipeline_info->pColorBlendState = blend_info;
	pipeline_info->pDynamicState = dynamic_info;
	pipeline_info->renderPass = handle_cast<::VkRenderPass>(render_pass);
	pipeline_info->subpass = 0;

	if(self.shprog)
	{
		pipeline_info->stageCount = self.shprog->n_stages;
		pipeline_info->pStages = reinterpret_cast<const VkPipelineShaderStageCreateInfo *>(self.shprog->creation_info.data());
		pipeline_info->layout = handle_cast<::VkPipelineLayout>(self.shprog->layout_handle);
	}

	if(self.vertex_setup)
	{
		self.vertex_setup->refresh();
		pipeline_info->pVertexInputState = reinterpret_cast<const VkPipelineVertexInputStateCreateInfo *>(self.vertex_setup->creation_info.data());
	}
}

uint64_t VulkanPipelineState::compute_descriptor_set_hash(unsigned index) const
{
	const PipelineState &self = *static_cast<const PipelineState *>(this);

	uint64_t result = hash<64>(0, 0);
	for(const PipelineState::BoundUniformBlock &b: self.uniform_blocks)
		if(b.used && b.binding>=0 && static_cast<unsigned>(b.binding>>20)==index)
		{
			result = hash_update<64>(result, b.binding);
			result = hash_update<64>(result, reinterpret_cast<uintptr_t>(b.block));
		}
	for(const PipelineState::BoundTexture &t: self.textures)
		if(t.used && (t.binding>>20)==index)
		{
			result = hash_update<64>(result, t.binding);
			result = hash_update<64>(result, reinterpret_cast<uintptr_t>(t.texture));
			result = hash_update<64>(result, reinterpret_cast<uintptr_t>(t.sampler));
		}

	return result;
}

VkDescriptorSetLayout VulkanPipelineState::get_descriptor_set_layout(unsigned index) const
{
	return static_cast<const PipelineState *>(this)->shprog->desc_set_layout_handles[index];
}

unsigned VulkanPipelineState::fill_descriptor_writes(unsigned index, vector<char> &buffer) const
{
	const PipelineState &self = *static_cast<const PipelineState *>(this);

	unsigned n_buffers = 0;
	for(const PipelineState::BoundUniformBlock &u: self.uniform_blocks)
		if(u.used && u.binding>=0 && static_cast<unsigned>(u.binding>>20)==index)
			++n_buffers;
	unsigned n_images = 0;
	for(const PipelineState::BoundTexture &t: self.textures)
		if(t.used && (t.binding>>20)==index)
			++n_images;
	unsigned n_writes = n_buffers+n_images;

	StructureBuilder sb(buffer, 3);
	VkWriteDescriptorSet *&writes = sb.add<VkWriteDescriptorSet>(n_writes);
	VkDescriptorBufferInfo *&buffers = sb.add<VkDescriptorBufferInfo>(n_buffers);
	VkDescriptorImageInfo *&images = sb.add<VkDescriptorImageInfo>(n_images);

	VkWriteDescriptorSet *write_ptr = writes;
	VkDescriptorBufferInfo *buffer_ptr = buffers;
	VkDescriptorImageInfo *image_ptr = images;

	for(const PipelineState::BoundUniformBlock &u: self.uniform_blocks)
		if(u.used && u.binding>=0 && static_cast<unsigned>(u.binding>>20)==index)
		{
			buffer_ptr->buffer = handle_cast<::VkBuffer>(u.block->get_buffer()->handle);
			buffer_ptr->offset = u.block->get_offset();
			buffer_ptr->range = u.block->get_data_size();

			write_ptr->sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
			write_ptr->dstBinding = u.binding&0xFFFFF;
			write_ptr->descriptorCount = 1;
			write_ptr->descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
			write_ptr->pBufferInfo = buffer_ptr;

			++buffer_ptr;
			++write_ptr;
		}

	for(const PipelineState::BoundTexture &t: self.textures)
		if(t.used && (t.binding>>20)==index)
		{
			image_ptr->sampler = handle_cast<::VkSampler>(t.sampler->handle);
			image_ptr->imageView = handle_cast<::VkImageView>(t.texture->view_handle);
			image_ptr->imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;

			write_ptr->sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
			write_ptr->dstBinding = t.binding&0xFFFFF;
			write_ptr->descriptorCount = 1;
			write_ptr->descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
			write_ptr->pImageInfo = image_ptr;

			++image_ptr;
			++write_ptr;
		}

	return n_writes;
}

void VulkanPipelineState::apply(VkCommandBuffer command_buffer) const
{
	const PipelineState &self = *static_cast<const PipelineState *>(this);
	const VulkanFunctions &vk = device.get_functions();

	vk.CmdBindPipeline(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, handle);
	if(const VertexSetup *vs = self.vertex_setup)
	{
		vk.CmdBindVertexBuffers(command_buffer, 0, vs->n_bindings, vs->buffers, vs->offsets);
		VkIndexType index_type = static_cast<VkIndexType>(get_vulkan_index_type(vs->get_index_type()));
		vk.CmdBindIndexBuffer(command_buffer, vs->get_index_buffer()->handle, 0, index_type);
	}

	if(!self.uniform_blocks.empty())
	{
		const PipelineState::BoundUniformBlock &first_block = self.uniform_blocks.front();
		if(first_block.used && first_block.binding==ReflectData::PUSH_CONSTANT)
		{
			const UniformBlock &pc_block = *first_block.block;
			vk.CmdPushConstants(command_buffer, self.shprog->layout_handle, self.shprog->stage_flags,
				pc_block.get_offset(), pc_block.get_data_size(), pc_block.get_data_pointer());
		}
	}

	vk.CmdBindDescriptorSets(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, self.shprog->layout_handle, 0, descriptor_set_handles.size(), descriptor_set_handles.data(), 0, 0);

	VkViewport viewport = { };
	if(self.viewport)
	{
		viewport.x = self.viewport->left;
		viewport.y = self.framebuffer->get_height()-(self.viewport->bottom+self.viewport->height);
		viewport.width = self.viewport->width;
		viewport.height = self.viewport->height;
	}
	else
	{
		viewport.x = 0;
		viewport.y = 0;
		viewport.width = self.framebuffer->get_width();
		viewport.height = self.framebuffer->get_height();
	}
	viewport.minDepth = 0.0f;
	viewport.maxDepth = 1.0f;
	vk.CmdSetViewport(command_buffer, 0, 1, &viewport);

	VkRect2D scissor = { };
	if(self.scissor)
	{
		scissor.offset.x = self.scissor->left;
		scissor.offset.y = self.framebuffer->get_height()-(self.scissor->bottom+self.scissor->height);
		scissor.extent.width = self.scissor->width;
		scissor.extent.height = self.scissor->height;
	}
	else
	{
		scissor.offset.x = 0;
		scissor.offset.y = 0;
		scissor.extent.width = self.framebuffer->get_width();
		scissor.extent.height = self.framebuffer->get_height();
	}
	vk.CmdSetScissor(command_buffer, 0, 1, &scissor);
}

} // namespace GL
} // namespace Msp
