#include "pixelformat.h"
#include "pixelformat_backend.h"
#include "vulkan.h"

using namespace std;

namespace {

static unsigned swizzle_orders[] =
{
	VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_G, VK_COMPONENT_SWIZZLE_B, VK_COMPONENT_SWIZZLE_A,
	VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_ONE,
	VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_G,
	VK_COMPONENT_SWIZZLE_B, VK_COMPONENT_SWIZZLE_G, VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_A,
	VK_COMPONENT_SWIZZLE_R, VK_COMPONENT_SWIZZLE_G, VK_COMPONENT_SWIZZLE_B, VK_COMPONENT_SWIZZLE_ONE
};

}

namespace Msp {
namespace GL {

ComponentSwizzle get_required_swizzle(PixelComponents comp)
{
	switch(comp)
	{
	case RGB: return RGBA_TO_RGB;
	case LUMINANCE: return R_TO_LUMINANCE;
	case LUMINANCE_ALPHA: return RG_TO_LUMINANCE_ALPHA;
	default: return NO_SWIZZLE;
	}
}

void require_pixelformat(PixelFormat)
{ }

unsigned get_vulkan_pixelformat(PixelFormat pf)
{
	switch(pf)
	{
	case R8: return VK_FORMAT_R8_UNORM;
	case R16F: return VK_FORMAT_R16_SFLOAT;
	case R32F: return VK_FORMAT_R32_SFLOAT;
	case RG8: return VK_FORMAT_R8G8_UNORM;
	case RG16F: return VK_FORMAT_R16G16_SFLOAT;
	case RG32F: return VK_FORMAT_R32G32_SFLOAT;
	case RGB8: return VK_FORMAT_R8G8B8_UNORM;
	case RGB16F: return VK_FORMAT_R16G16B16_SFLOAT;
	case RGB32F: return VK_FORMAT_R32G32B32_SFLOAT;
	case RGBA8: return VK_FORMAT_R8G8B8A8_UNORM;
	case RGBA16F: return VK_FORMAT_R16G16B16A16_SFLOAT;
	case RGBA32F: return VK_FORMAT_R32G32B32A32_SFLOAT;
	case SRGB8: return VK_FORMAT_R8G8B8_SRGB;
	case SRGB8_ALPHA8: return VK_FORMAT_R8G8B8A8_SRGB;
	case BGR8: return VK_FORMAT_B8G8R8_UNORM;
	case BGRA8: return VK_FORMAT_B8G8R8A8_UNORM;
	case SBGR8: return VK_FORMAT_B8G8R8_SRGB;
	case SBGR8_ALPHA8: return VK_FORMAT_B8G8R8A8_SRGB;
	case DEPTH_COMPONENT16: return VK_FORMAT_D16_UNORM;
	case DEPTH_COMPONENT24: return VK_FORMAT_X8_D24_UNORM_PACK32;
	case DEPTH_COMPONENT32F: return VK_FORMAT_D32_SFLOAT;
	case STENCIL_INDEX8: return VK_FORMAT_S8_UINT;
	default: throw invalid_argument("get_vulkan_pixelformat");
	}
}

unsigned get_vulkan_aspect(PixelComponents comp)
{
	switch(comp)
	{
	case DEPTH_COMPONENT: return VK_IMAGE_ASPECT_DEPTH_BIT;
	case STENCIL_INDEX: return VK_IMAGE_ASPECT_STENCIL_BIT;
	default: return VK_IMAGE_ASPECT_COLOR_BIT;
	}
}

unsigned get_vulkan_attachment_layout(PixelComponents comp)
{
	switch(comp)
	{
	case DEPTH_COMPONENT:
	case STENCIL_INDEX: return VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
	default: return VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
	}
}

PixelFormat pixelformat_from_vulkan(unsigned vkf)
{
	switch(vkf)
	{
	case VK_FORMAT_R8_UNORM: return R8;
	case VK_FORMAT_R16_SFLOAT: return R16F;
	case VK_FORMAT_R32_SFLOAT: return R32F;
	case VK_FORMAT_R8G8_UNORM: return RG8;
	case VK_FORMAT_R16G16_SFLOAT: return RG16F;
	case VK_FORMAT_R32G32_SFLOAT: return RG32F;
	case VK_FORMAT_R8G8B8_UNORM: return RGB8;
	case VK_FORMAT_R16G16B16_SFLOAT: return RGB16F;
	case VK_FORMAT_R32G32B32_SFLOAT: return RGB32F;
	case VK_FORMAT_R8G8B8A8_UNORM: return RGBA8;
	case VK_FORMAT_R16G16B16A16_SFLOAT: return RGBA16F;
	case VK_FORMAT_R32G32B32A32_SFLOAT: return RGBA32F;
	case VK_FORMAT_R8G8B8_SRGB: return SRGB8;
	case VK_FORMAT_R8G8B8A8_SRGB: return SRGB8_ALPHA8;
	case VK_FORMAT_B8G8R8_UNORM: return BGR8;
	case VK_FORMAT_B8G8R8A8_UNORM: return BGRA8;
	case VK_FORMAT_B8G8R8_SRGB: return SBGR8;
	case VK_FORMAT_B8G8R8A8_SRGB: return SBGR8_ALPHA8;
	case VK_FORMAT_D16_UNORM: return DEPTH_COMPONENT16;
	case VK_FORMAT_X8_D24_UNORM_PACK32: return DEPTH_COMPONENT24;
	case VK_FORMAT_D32_SFLOAT: return DEPTH_COMPONENT32F;
	case VK_FORMAT_S8_UINT: return STENCIL_INDEX8;
	default: throw invalid_argument("pixelformat_from_vulkan");
	}
}

const unsigned *get_vulkan_swizzle(ComponentSwizzle swiz)
{
	return swizzle_orders+4*swiz;
}

} // namespace GL
} // namespace Msp
