#include <msp/core/algorithm.h>
#include "device.h"
#include "transferqueue.h"
#include "vulkan.h"

using namespace std;

namespace Msp {
namespace GL {

TransferQueue::TransferQueue(Device &d):
	device(d)
{ }

TransferQueue::PendingTransfer &TransferQueue::prepare_transfer(size_t size)
{
	auto i = find_if(buffers, [size](const StagingBuffer &b){ return b.used+size<=b.size; });
	if(i==buffers.end())
	{
		buffers.emplace_back(device, max(default_buffer_size, size));
		i = prev(buffers.end());
	}

	PendingTransfer transfer;
	transfer.buffer_index = distance(buffers.begin(), i);
	transfer.offset = i->used;
	transfer.size = size;
	transfers.push_back(transfer);

	i->used += size;

	return transfers.back();
}

void TransferQueue::dispatch_transfers(VkCommandBuffer command_buffer)
{
	if(!transfers.empty())
		device.get_synchronizer().barrier(command_buffer);

	for(const PendingTransfer &t: transfers)
	{
		VkBuffer buffer = buffers[t.buffer_index].buffer;
		t.callback(command_buffer, buffer, t.offset);
	}

	transfers.clear();
}


TransferQueue::StagingBuffer::StagingBuffer(Device &d, size_t s):
	device(d),
	size(s)
{
	const VulkanFunctions &vk = device.get_functions();

	VkBufferCreateInfo buffer_info = { };
	buffer_info.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
	buffer_info.size = size;
	buffer_info.usage = VK_BUFFER_USAGE_TRANSFER_SRC_BIT;
	buffer_info.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

	vk.CreateBuffer(buffer_info, buffer);

	MemoryAllocator &allocator = device.get_allocator();
	memory_id = allocator.allocate(buffer, STAGING_MEMORY);
	mapped_address = allocator.map(memory_id, 0, size);
}

TransferQueue::StagingBuffer::StagingBuffer(StagingBuffer &&other):
	device(other.device),
	buffer(other.buffer),
	memory_id(other.memory_id),
	size(other.size),
	used(other.used),
	mapped_address(other.mapped_address)
{
	other.buffer = 0;
	other.memory_id = 0;
	other.mapped_address = 0;
}

TransferQueue::StagingBuffer::~StagingBuffer()
{
	const VulkanFunctions &vk = device.get_functions();
	MemoryAllocator &allocator = device.get_allocator();

	if(mapped_address)
	{
		allocator.unmap(mapped_address);
		allocator.release(memory_id);
	}
	if(buffer)
		vk.DestroyBuffer(buffer);
}

} // namespace GL
} // namespace Msp
